-module(eAcs).

-export([
   matchSw/1                 %% 返回匹配的敏感词列表
   , isHasSw/1               %% 检查是否包含敏感词
   , replaceSw/1             %% 替换敏感词
   , strSize/2               %% 获取utf8字符串的长度
]).

%% state 0 is the root node
%% Goto:  State -> #{Word -> State}
%% failOut: State -> {FailState, BinStr}
%% *************************************** matchSw start ***************************************************************
-spec matchSw(BinStr :: binary()) -> [{StartIndex :: integer(), EndIndex :: integer(), Pattern :: binary()}].
matchSw(BinStr) ->
   doMatchMs(BinStr, 0, _Index = 1, _MatchList = []).

doMatchMs(<<>>, _, _Index, MatchList) ->
   MatchList;
doMatchMs(<<Word/utf8, Tail/binary>>, State, Index, MatchList) ->
   case acsSpw:getSpw(Word) of
      true ->
         %% 特殊字符跳过 且Index 不加 1
         doMatchMs(Tail, State, Index, MatchList);
      _ ->
         {NewState, NewMatchList} = matchWordMs(Word, State, Index, MatchList),
         doMatchMs(Tail, NewState, Index + 1, NewMatchList)
   end.

matchWordMs(Word, State, Index, MatchList) ->
   Node = acsTree:goto(State),
   case Node of
      undefined ->
         case State of
            0 ->
               {State, MatchList};
            _ ->
               {NextState, _} = acsTree:failOut(State),
               matchWordMs(Word, NextState, Index, MatchList)
         end;
      _ ->
         case Node of
            #{Word := NextState} ->
               NewMatchList = getOutputMs(NextState, Index, MatchList),
               {NextState, NewMatchList};
            _ ->
               case State of
                  0 ->
                     {State, MatchList};
                  _ ->
                     {NextState, _} = acsTree:failOut(State),
                     matchWordMs(Word, NextState, Index, MatchList)
               end
         end
   end.

getOutputMs(0, _Index, MatchList) ->
   MatchList;
getOutputMs(State, Index, MatchList) ->
   {FailState, Pattern} = acsTree:failOut(State),
   case Pattern of
      undefined ->
         getOutputMs(FailState, Index, MatchList);
      _ ->
         NewMatchList = [{Index - Pattern + 1, Pattern} | MatchList],
         getOutputMs(FailState, Index, NewMatchList)
   end.

%% *************************************** matchSw end   ***************************************************************
%% *************************************** isHasSw start ***************************************************************
-spec isHasSw(BinStr :: binary()) -> boolean().
isHasSw(BinStr) ->
   doMatchIs(BinStr, 0).

doMatchIs(<<>>, _) ->
   false;
doMatchIs(<<Word/utf8, Tail/binary>>, State) ->
   case acsSpw:getSpw(Word) of
      true ->
         %% 特殊字符跳过
         doMatchIs(Tail, State);
      _ ->
         case matchWordIs(Word, State) of
            true ->
               true;
            NewState ->
               doMatchIs(Tail, NewState)
         end
   end.

matchWordIs(Word, State) ->
   Node = acsTree:goto(State),
   case Node of
      undefined ->
         case State of
            0 ->
               State;
            _ ->
               {NextState, _} = acsTree:failOut(State),
               matchWordIs(Word, NextState)
         end;
      _ ->
         case Node of
            #{Word := NextState} ->
               case getOutputIs(NextState) of
                  false ->
                     NextState;
                  _ ->
                     true
               end;
            _ ->
               case State of
                  0 ->
                     State;
                  _ ->
                     {NextState, _} = acsTree:failOut(State),
                     matchWordIs(Word, NextState)
               end
         end
   end.

getOutputIs(0) ->
   false;
getOutputIs(State) ->
   {FailState, Pattern} = acsTree:failOut(State),
   case Pattern of
      undefined ->
         getOutputIs(FailState);
      _ ->
         true
   end.
%% *************************************** matchSw end   ***************************************************************
%% *************************************** replaceSw start *************************************************************
replaceSw(BinStr) ->
   MatchBIMWs = doMatchRs(BinStr, _TotalSize = byte_size(BinStr) - 1, _Index = 1, _State = 0, _MatchList = []),
   doReplaceSw(lists:reverse(MatchBIMWs), 0, BinStr, <<>>).

%% 从前往后替换
doReplaceSw([], _BinStr, _StartPos, Acc) ->
   iolist_to_binary(Acc);
doReplaceSw([{CurByteIndex, MatchWordCnt, _CurWordIndex} | MatchBIMWs], _BinStr, StartPos, Acc) ->
   iolist_to_binary(Acc).

%% 递归处理判断最新的匹配是否包含或者连接了历史匹配的{CurByteIndex, MatchWordCnt, CurWordIndex}
dealMatchList([], CurByteIndex, MatchWordCnt, CurWordIndex) ->
   [{CurByteIndex, MatchWordCnt, CurWordIndex}];
dealMatchList([{_OldByteIndex, OldMatchWordCnt, OldWordIndex} | LeftMatchList] = OldMatchList, CurByteIndex, MatchWordCnt, CurWordIndex) ->
   CurStartIndex = CurWordIndex - MatchWordCnt,
   OldStartIndex = OldWordIndex - OldMatchWordCnt,
   if
      CurStartIndex > OldWordIndex + 1 ->
         [{CurByteIndex, MatchWordCnt, CurWordIndex} | OldMatchList];
      CurStartIndex >= OldStartIndex ->
         [{CurByteIndex, CurWordIndex - OldStartIndex, CurWordIndex} | LeftMatchList];
      true ->
         dealMatchList(LeftMatchList, CurByteIndex, MatchWordCnt, CurWordIndex)
   end.

doMatchRs(<<>>, _TotalSize, _CurIndex, _State, MatchList) ->
   MatchList;
doMatchRs(<<Word/utf8, Tail/binary>>, TotalSize, CurIndex, State, MatchList) ->
   case acsSpw:getSpw(Word) of
      true ->
         %% 特殊字符跳过 且CurIndex 不加 1
         doMatchRs(Tail, TotalSize, CurIndex, State, MatchList);
      _ ->
         {NewState, MatchCnt} = matchWordRs(Word, State, 0),
         case MatchCnt of
            0 ->
               doMatchRs(Tail, TotalSize, CurIndex + 1, NewState, MatchList);
            _ ->
               LeftSize = byte_size(Tail),
               NewMatchList = dealMatchList(MatchList, TotalSize - LeftSize, MatchCnt, CurIndex),
               doMatchRs(Tail, TotalSize, CurIndex + 1, NewState, NewMatchList)
         end
   end.

matchWordRs(Word, State, MatchCnt) ->
   Node = acsTree:goto(State),
   case Node of
      undefined ->
         case State of
            0 ->
               {State, MatchCnt};
            _ ->
               {NextState, _} = acsTree:failOut(State),
               matchWordRs(Word, NextState, MatchCnt)
         end;
      _ ->
         case Node of
            #{Word := NextState} ->
               NewMatchCnt = getOutputRs(NextState, MatchCnt),
               {NextState, NewMatchCnt};
            _ ->
               case State of
                  0 ->
                     {State, MatchCnt};
                  _ ->
                     {NextState, _} = acsTree:failOut(State),
                     matchWordRs(Word, NextState, MatchCnt)
               end
         end
   end.

%% 获取当前字符最大匹配数
getOutputRs(0, MatchCnt) ->
   MatchCnt;
getOutputRs(State, MatchCnt) ->
   {FailState, Pattern} = acsTree:failOut(State),
   case Pattern of
      undefined ->
         getOutputRs(FailState, MatchCnt);
      _ ->
         case Pattern > MatchCnt of
            true ->
               getOutputRs(FailState, Pattern);
            _ ->
               getOutputRs(FailState, MatchCnt)
         end
   end.
% *************************************** replaceSw end   *************************************************************

strSize(<<>>, Cnt) ->
   Cnt;
strSize(<<_Word/utf8, Left/binary>>, Cnt) ->
   strSize(Left, Cnt + 1).